#!/usr/bin/env python
#
# The Kepler grammar using Lark parser library
#
# Author:  Raul E. Monti
# Date:    15.10.2021
# License: GPLv3
#
from lark import Lark
from lark import Transformer
from lark import Visitor
from lark import Tree
from lark import Token
from lark.visitors import Discard
from iosa import IosaApi
from numpy import lcm as nplcm
from itertools import combinations
from pyeda.inter import expr
from pyeda.boolalg.expr import Variable as PyedaVariable

#
# Constants for easy setup
#

FIG_MAX_IVAL_PAND = 3       # max value assigned by FIG to the if of a pand

TYPE_TO_PREFIX = {
    "and_gate"  :"AND_",
    "or_gate"   :"OR",
    "pand_gate" :"PAND_",
    "vot_gate"  :"VOT_",
    "spare_gate":"SPARE_",
    "be"        :"BE_",
    "sbe"       :"SBE_",
    "repair_box":"RBOX_",
    "fdep_gate" :"FDEP_"
}

#
# Kepler grammar using Lark parser
#

kepler_parser = Lark(r"""
    toplevel : "toplevel" name ";" (kepler | galileo)
             
    kepler   : (gate
             | be
             | sbe)+
    gate     : name "or"    name+ ";"        -> or_gate
             | name "and"   name+ ";"        -> and_gate
             | name  INT"of"INT name+ ";"    -> vot_gate
             | name "pand"  name+ ";"        -> pand_gate
             | name "fdep"  name+ ";"        -> fdep_gate
             | name "spare" name+ ";"        -> spare_gate
             | name "rbox"  RMODE name+ ";"  -> repair_box
    be       : name fail ";"
             | name (fail rep | rep fail) ";"
    sbe      : name (dorm fail | fail dorm) ";"
             | name (dorm fail rep | dorm rep fail | fail dorm rep | fail rep dorm | rep fail dorm | rep dorm fail) ";"
    RMODE    : "prio" | "fcfs" | "rand"
    fail    : "fail"   "~" dist            
    rep     : "repair" "~" dist
    dorm    : "dorm"   "~" dist
   
   
    galileo : (ggate
             | gbe)+
    ggate    : name "seq" name+ ";"          -> galileo_seq_gate
             | name "csp" name+ ";"          -> galileo_csp_gate
             | name "wsp" name+ ";"          -> galileo_wsp_gate
             | name "hsp" name+ ";"          -> galileo_hsp_gate
    gbe      : name gfail ";"                -> galileo_be
             | name gfail gdorm ";"          -> galileo_be
             | name gdorm gfail ";"          -> galileo_be
    gfail    : "lambda" "=" NUMBER
             | "prob"   "=" PROB
             | "res"    "=" PROB
             | "rate"   "=" NUMBER "shape" "=" NUMBER
             | "mean"   "=" REAL  "stddev" "=" NUMBER
    gdorm    : "dorm"   "=" PROB


    ?name     : "\"" CNAME "\""
    dist     : "exponential" "(" NUMBER ")"             -> exponential_dist
             | "erlang"      "(" INT   "," NUMBER  ")"  -> erlang_dist
             | "uniform"     "(" REAL  "," REAL    ")"  -> uniform_dist
             | "normal"      "(" REAL  "," NUMBER  ")"  -> normal_dist
             | "lognormal"   "(" REAL  "," NUMBER  ")"  -> lognormal_dist
             | "gamma"       "(" NUMBER "," NUMBER ")"  -> gamma_dist
             | "weibull"     "(" NUMBER "," NUMBER ")"  -> weibull_dist
             | "rayleigh"    "(" NUMBER ")"             -> rayleigh_dist
             | "dirac"       "(" NUMBER ")"             -> dirac_dist
    PROB     : "0" "." INT | "0" | "1"
    REAL     : ["+" | "-"] NUMBER
    SGNINT   : ["+" | "-"] INT

    COMMENT  : /\/\/.*/ | /\/\*.*\*\//
    
    %import common.CNAME
    %import common.INT
    %import common.DIGIT
    %import common.WS
    %ignore COMMENT
    %import common.FLOAT
    %import common.NUMBER
    %ignore WS
    
    """, start = "toplevel")


def get_n_inputs(tree):
    """ Return the number of inputs for each gate of RFT
    """
    n = len(tree.children)
    if tree.data == "vot_gate":
        return n - 3
    elif tree.data == "repair_box":
        return n - 2
    else:
        return n - 1

def div(d, dd):
    """ Simplifies away divisions by 1 and some other cases. 
    """
    if str(dd) == "0":
        return "ERROR: div by 0"
    elif str(dd) == "1":
        return "{}".format(d)
    elif str(d) == str(dd):
        return "1"
    else:
        res = "({}/{})".format(d,dd) 
        try:
            res = int(eval(res))
        except:
            pass
        return res
        
def prod(m, mm):
    """ Simplifies away multiplications by 1. 
    """
    if str(m) == "1":
        return str(mm)
    elif str(mm) == "1":
        return str(m)
    elif str(m) == "-1":
        return "-{}".format(mm)
    elif str(mm) == "-1":
        return "-{}".format(m)
    else:
        res = ("({}*{})").format(m,mm)
        try:
            res = int(eval(res))
        except:
            pass
        return res

        
def mx(ls):
    if len(ls) > 1:
        result = "max(" +  ','.join(ls) + ')'
    else:
        result = ls[0]
    return result

    
class TransformFDEP(Visitor):
    
    def __init__(self, tree):
        self.fdep_map = {}
        self.find_fdeps(tree)
    
    def find_fdeps(self,tree):
        for e in tree.find_data("fdep_gate"):
            _name, _main, *_ch = e.children
            _name = _name.value
            _main = _main.value
            for be in _ch:
                self.fdep_map.setdefault(be.value, []).append(_main)

    def make_or_gate(self, name, *ch):
        _children = []
        for c in [name] + list(ch):
            _t = Token('CNAME', str(c))
            _children.append(_t)
        return Tree('or_gate',  _children)
    
    def name_for_or_gate(self, be_name):
        return "OR_FROM_FDEP_" + be_name
    
    def replace_be_by_or(self, children):
        for i,e in enumerate(children):
            try:
                _deps = self.fdep_map[str(e)]
                tree.children[i] = Token('CNAME', self.name_for_or_gate(e.value))
            except:
                pass
    
    def and_gate(self, tree):
        self.replace_be_by_or(tree.children[1::])

    def or_gate(self, tree):
        self.replace_be_by_or(tree.children[1::])
        
    def vot_gate(self, tree):
        self.replace_be_by_or(tree.children[3::])

    def pand_gate(self, tree):
        self.replace_be_by_or(tree.children[1::])
        
    def kepler(self, tree):
        for k, v in self.fdep_map.items():
            name = self.name_for_or_gate(str(k))
            gate = self.make_or_gate(name, *(v+[k]))
            tree.children.append(gate)


class CompilePass(Transformer):
    """ Final pass, should be able to transform
        a processed DFT into a valid IOSA model
        using the IOSA module API.
    """
    
    def __init__(self, tree):
        self.api = IosaApi()
        self.sbe_2_gates = {}
        self.fill_gates(tree)
        
    def fill_gates(self, tree):
        for e in tree.find_data("spare_gate"):
            for sbe in e.children[2::]:
                self.sbe_2_gates.setdefault(sbe, []).append(e.children[0])
    
    def toplevel(self, args) :
        return args[1]

    def be(self, args):
        d = {}
        for e in args[1::]:
            d[e.data] = e.children[0]
        return self.api.make_be(args[0],
            d['fail'],
            d['rep'] if len(args) > 2 else "")
    
    def sbe(self, args):
        _gates = [x.value for x in self.sbe_2_gates[args[0]]]
        d = {}
        for e in args[1::]:
            d[e.data] = e.children[0]
        return self.api.make_sbe(args[0],
            d['dorm'],
            d['fail'],
            d['rep'] if len(args) > 3 else "",
            _gates)
        
    def and_gate(self, args):
        return self.api.make_and(args[0], args[1:])

    def or_gate(self, args):
        return self.api.make_or(args[0], args[1:])
    
    def pand_gate(self, args):
        return self.api.make_pand(args[0], args[1:])
        
    def vot_gate(self, args):
        return self.api.make_vot(args[0], args[1], args[3::])
        
    def fdep_gate(self, args):
        raise Discard
        
    def spare_gate(self, args):
        return self.api.make_spare(args[0], args[1::])
    
    def repair_box(self, args):
        return self.api.make_rbox(args[0], args[1], args[2:])
    
    def uniform_dist(self, args):
        return self.api.make_dist_uniform(args[0], args[1])
        
    def exponential_dist(self, args):
        return self.api.make_dist_exponential(args[0])

    def weibull_dist(self, args):
        return self.api.make_dist_weibull(args[0], args[1])

    def lognormal_dist(self, args):
        return self.api.make_dist_lognormal(args[0], args[1])

    def normal_dist(self, args):
        return self.api.make_dist_normal(args[0], args[1])

    def gamma_dist(self, args):
        return self.api.make_dist_gamma(args[0], args[1])        

    def rayleigh_dist(self, args):
        return self.api.make_dist_rayleigh(args[0])

    def erlang_dist(self, args):
        return self.api.make_dist_erlang(args[0], args[1])
        
    def dirac_dist(self, args):
        return self.api.make_dist_dirac(args[0])
        
    def name(self, args):
        return args[0]
        
    def CNAME(self, name):
        return name.value
        
    def RMODE(self, mode):
        return mode.value


class Kepler2Tree(Visitor):
    
    class NodesNameDir(Visitor):
        """ Visits the a Parsed Kepler tree and builds a 
        dictionary where the keys are the names of the 
        nodes and the values are the actual Tree structures 
        representing the nodes. With this map you can, for
        instance, easily build a tree for further analysis.
    
        The resulting map is placed in self.dir and you
        can access it with get_dir().
        """
    
        def __init__(self):
            self.dir = {}
            return
    
        def get_dir(self):
            return self.dir
    
        def and_gate(self, tree):
            self.dir[tree.children[0].value] = tree
        
        def or_gate(self, tree):
            self.dir[tree.children[0].value] = tree

        def pand_gate(self, tree):
            self.dir[tree.children[0].value] = tree

        def vot_gate(self, tree):
            self.dir[tree.children[0].value] = tree
    
        def be(self, tree):
            self.dir[tree.children[0].value] = tree
    
        def sbe(self, tree):
            self.dir[tree.children[0].value] = tree

        def spare_gate(self, tree):
            self.dir[tree.children[0].value] = tree
    
    def __init__(self, k_tree):
        __dir_pass = self.NodesNameDir()
        __dir_pass.visit(k_tree)
        self.dir = __dir_pass.get_dir()
        
    def toplevel(self, tree):
        tree.children = [self.dir[tree.children[0].value]]
    
    def and_gate(self, tree):
        for i, e in enumerate(tree.children):
            if(i > 0):
                tree.children[i] = self.dir[e.value]
        
    def or_gate(self ,tree):
        for i, e in enumerate(tree.children):
            if(i > 0):
                tree.children[i] = self.dir[e.value]

    def pand_gate(self ,tree):
        for i, e in enumerate(tree.children):
            if(i > 0):
                tree.children[i] = self.dir[e.value]

    def vot_gate(self, tree):
        for i, e in enumerate(tree.children):
            if(i > 2):
                tree.children[i] = self.dir[e.value]

    def spare_gate(self, tree):
        for i, e in enumerate(tree.children):
            if(i > 0):
                tree.children[i] = self.dir[e.value]


class Tree2StructuralIfun(Visitor):
    """
    Defined in Budde et all. 'Rare event simulation for 
    non-Markovian repairable fault trees' (Table 1)
    """
        
    def lcm(self, *args):
        return nplcm.reduce(list(args))
    
    def lcmv(self, *args):
        return str(self.lcm(*[self.maxi(x) for x in args]))

    # TODO        
    def mini(self, tree):
        """Return the minimum structural importance value of a node"""
        return
    
    def maxi(self, tree):
        """ 
        Returns the maximum structural importance value of a node.
        Could optimise saving pre-calculated values in tree nodes, 
        but code is uglier and maybe not needed given the expected
        small size of trees.
        """
        result = -1
        
        try:
            tree.maxi
        except:
            if tree.data in {"be", "sbe"}:
                tree.maxi = 1   
            elif tree.data == "and_gate":
                lcm_v = self.lcm(*[self.maxi(x) for x in tree.children[1::]])
                tree.maxi = lcm_v * (len(tree.children) - 1)
            elif tree.data == "or_gate":
                tree.maxi = self.lcm(*[self.maxi(x) for x in tree.children[1::]])
            elif tree.data == "vot_gate":
                lcm_v = self.lcm(*[self.maxi(x) for x in tree.children[3::]])
                tree.maxi = lcm_v * int(tree.children[1])
            elif tree.data == "pand_gate":
                lcm_v = self.lcm(*[self.maxi(x) for x in tree.children[1::]])
                tree.maxi = lcm_v * 2
            elif tree.data == "spare_gate":
                lcm_v = self.lcm(*[self.maxi(x) for x in tree.children[1::]])
                tree.maxi = lcm_v * (len(tree.children) - 1)
        return tree.maxi 

    def toplevel(self, tree):
        tree.ifun = tree.children[0].ifun
        tree.maxi = self.maxi(tree.children[0])
        
    def be(self, tree):
        tree.ifun = "BE_" + tree.children[0]
        
    def sbe(self, tree):
        tree.ifun = "SBE_" + tree.children[0]

    def div(self, d, dd):
        """ Simplifies away divisions by 1 and some other cases. """
        if str(dd) == "0":
            return "ERROR: div by 0"
        elif str(dd) == "1":
            return "{}".format(d)
        elif str(d) == str(dd):
            return "1"
        else:
            res = "({}/{})".format(d,dd) 
            try:
                res = int(eval(res))
            except:
                pass
            return res
            
    def prod(self, m, mm):
        """ Simplifies away multiplications by 1. """
        if str(m) == "1":
            return str(mm)
        elif str(mm) == "1":
            return str(m)
        elif str(m) == "-1":
            return "-{}".format(mm)
        elif str(mm) == "-1":
            return "-{}".format(m)
        else:
            res = ("({}*{})").format(m,mm)
            try:
                res = int(eval(res))
            except:
                pass
            return res

    def and_gate(self, tree):
        _cs = tree.children[1::]
        _lcmv = self.lcmv(*_cs)
        _sum = []
        for e in _cs:
            _scale = div(_lcmv, self.maxi(e))
            _sum.append(prod(_scale, e.ifun))
        _sum = "(" + "+".join(_sum) + ")"
        tree.ifun = _sum

    def or_gate(self, tree):
        _cs = tree.children[1::]
        _lcmv = self.lcmv(*_cs)
        _max = []
        for e in _cs:
            _scale = div(_lcmv, self.maxi(e))
            _max.append(prod(_scale, e.ifun))
        _max = "max(" + ",".join(_max) + ")"
        tree.ifun = _max

    def pand_gate(self, tree):
        _cs = tree.children
        _name = _cs[0]
        _lcmv = self.lcmv(*_cs[1::])
        _scalel = div(_lcmv, self.maxi(_cs[1]))
        _suml = prod(_scalel, _cs[1].ifun)
        _scaler = div(_lcmv, self.maxi(_cs[2]))
        _scif = prod(_scaler,_cs[2].ifun)
        _sumr = "({}>0?{}:-{})".format("PAND_"+_name, _scif, _scif) # >0 based on how FIG computes local ifun of pand
        _sum = "({}+{})".format(_suml, _sumr)
        _fig_max = FIG_MAX_IVAL_PAND
        _out = "({}=={})".format("PAND_"+_name, _fig_max)
        _max = "max({},{}?{}:0)".format(_sum, _out, prod(2, _lcmv))
        tree.ifun = _max

    def vot_gate(self, tree):
        _cs = tree.children
        _count = _cs[1]
        _from = _cs[2]
        _cs = _cs[3::]
        _lcmv = self.lcmv(*_cs)
        _prod = [prod(div(_lcmv,self.maxi(x)),x.ifun) for x in _cs]
        _max = "summax({},{})".format(_count, ','.join(_prod))
        tree.ifun = _max

    def spare_gate(self, tree):
        _fig_max = 1 + 2 *(len(tree.children)-2) # max value assigned by FIG to the if of a spare
        _cs = tree.children
        _name = _cs[0]
        _cs = _cs[1::]
        _lcmv = self.lcmv(*_cs)
        _sum = [prod(div(_lcmv, self.maxi(e)),e.ifun) for e in _cs]
        _sum = "(" + "+".join(_sum) + ")"
        _st = "({}=={}?1:0)".format("SPARE_"+_name, _fig_max)
        _max = "max({},{})".format(_sum, prod(_st, prod(len(_cs), _lcmv)))
        tree.ifun = _max


class Tree2MCSIfun(Visitor):
        
    def ifun(self, tree):
        _mcs = self.get_mcs(tree)
        _max = max([len(x) for x in _mcs])
        _mcs = ['(' + '+'.join(x) + ')' for x in _mcs]
        _ifun = "max(" + ','.join(_mcs) + ")"
        _ifun = mx(_mcs)
        return _ifun, 0, _max
        
    def ifun_n(self, tree):
        _mcs = self.get_mcs(tree)
        _count = [len(x) for x in _mcs]
        _lcm = self.lcm(*_count)
        _max = _lcm
        _sum = ['(' + '+'.join(x) + ')' for x in _mcs]
        _prod = [prod(div(_lcm, _count[i]), _sum[i]) for i in range(0,len(_sum))]
        _ifun = mx(_prod)
        return _ifun, 0, _max
        
    def get_mcs(self, tree):
        self.visit(tree)
        _flat = tree.ifun
        _dnf = expr(_flat).to_dnf()
        return self.mcs_from_dnf(_dnf)
        
    def mcs_from_dnf(self, _dnf):
        result = []
        mcs = []
        for e in _dnf.iter_dfs():
            if isinstance(e, PyedaVariable):
                mcs.append(str(e))
            elif mcs:
                result.append(mcs)
                mcs = []
            else:
                pass
        return result
  
    def maxi(self, tree):
        pass
        
    def lcm(self, *args):
        return nplcm.reduce(list(args))
    
    def toplevel(self, tree):
        tree.ifun = tree.children[0].ifun
        
    def be(self, tree):
        tree.ifun = "BE_"+tree.children[0]
        
    def sbe(self, tree):
        tree.ifun = "SBE_"+tree.children[0]

    def and_gate(self, tree):
        tree.ifun = '(' + '&'.join([c.ifun for c in tree.children[1::]]) + ')'

    def or_gate(self, tree):
        tree.ifun = '(' + '|'.join([c.ifun for c in tree.children[1::]]) + ')'

    def pand_gate(self, tree):
        raise Exception("Models with PAND gates are not amenable to MCS IFuns.")

    def vot_gate(self, tree):
        comb = combinations([e.ifun for e in tree.children[3::]], int(tree.children[1]))
        tree.ifun = '(' + '|'.join(['&'.join(c) for c in comb ]) + ')'

    def spare_gate(self, tree):
        tree.ifun = '(' + '&'.join([c.ifun for c in tree.children[1::]]) + ')'



class Kepler2IOSA():
    
    def __init__(self):
        return
        
    def compile(self, input):
        """
        @param <input>: a string with a kepler/galileo valid DFT description
        @returns: a string with a IOSA model of <input>
        """
        tree = kepler_parser.parse(input)
        # tree = Rename(tree).visit(tree)
        tree = TransformFDEP(tree).visit(tree)
        iosa = CompilePass(tree).transform(tree)
        result = '\n\n'.join([str(x) for x in iosa.children])
        return result
        
    def ifun_structural(self, input):
        """ 
        @param <input>: a string with a kepler/galileo valid DFT description
        """
        tree = kepler_parser.parse(input)
        # tree = Rename(tree).visit(tree)
        tree = TransformFDEP(tree).visit(tree)
        tree = Kepler2Tree(tree).visit(tree)
        ifun = Tree2StructuralIfun().visit(tree)    
        return ifun.ifun, 0, ifun.maxi
        
    def ifun_mcs(self, input):
        tree = kepler_parser.parse(input)
        # tree = Rename(tree).visit(tree)
        tree = TransformFDEP(tree).visit(tree)
        tree = Kepler2Tree(tree).visit(tree)
        ifun = Tree2MCSIfun().ifun(tree)    
        # return ifun.ifun, 0, ifun.maxi
        return ifun

    def ifun_mcsn(self, input):
        tree = kepler_parser.parse(input)
        # tree = Rename(tree).visit(tree)
        tree = TransformFDEP(tree).visit(tree)
        tree = Kepler2Tree(tree).visit(tree)
        ifun = Tree2MCSIfun().ifun_n(tree)    
        # return ifun.ifun, 0, ifun.maxi
        return ifun
        
    def top_level_event(self, input):
        """ Return the characteristic state formula for the top
            level event module the top level gate.
            @param <input>: a string with a kepler/galileo valid DFT description
        """
        tree = kepler_parser.parse(input)
        tl_node = next(tree.find_data("toplevel"))
        tl_gate_name = tl_node.children[0]
        tl_gate = next(tree.find_pred(lambda t: t.children[0] == tl_gate_name))
        tl_type = tl_gate.data
        tl_prefix = TYPE_TO_PREFIX[tl_type]
        n = get_n_inputs(tl_gate)
        k = tl_gate.children[1] if tl_type == "vot_gate" else 0
        return IosaApi().fail_condition(tl_gate_name, tl_type, n, k)
        
      
        
if __name__ == "__main__":
    
    kepler = '''toplevel "AND1"; 
    "AND1" and "B1" "OR1"; 
    "B1" fail~exponential(1.2); 
    "OR1" or "B2" "PAND1"; 
    "B2" fail~exponential(2.2); 
    "PAND1" pand "B3" "V1"; 
    "B3" fail~exponential(3.2);
    "V1" 3 of 4 "B4" "B5" "B6" "SP1";
    "B4" fail~exponential(4.2);
    "B5" fail~exponential(5.2);
    "B6" fail~exponential(6.2);
    "SP1" spare "B7" "SBE1" "SBE2";
    "B7" fail~exponential(7.2);
    "SBE1" dorm~exponential(7.2) fail~exponential(7.2);
    "SBE2" dorm~exponential(7.2) fail~exponential(7.2);
    '''

    parsed = kepler_parser.parse(kepler)
    print(parsed.pretty())
    
    print(";".join([str(x) for x in Kepler2IOSA().ifun_structural(kepler)]))
    
    hecs = '''
    toplevel "HECS";
    "HECS" or "IF" "MEM" "BUS" "PROC";

    "IF" or "SW" "HW";
    "SW"  fail~exponential(1.2);
    "HW"  fail~exponential(1.2);

    "MEM" 3of5 "M1" "M2" "M3" "M4" "M5";

    "MI3" and "MI1" "MI2";
    "MI2" fail~exponential(1.2);
    "MI1" fail~exponential(1.2);
    "M1"  fail~exponential(1.2);
    "M2"  fail~exponential(1.2);
    "M3"  fail~exponential(1.2);
    "M4"  fail~exponential(1.2);
    "M5"  fail~exponential(1.2);

    "BUS" and "B1" "B2" "B3" "B4" "B5" "B6" "B7" "B8";
    "B1"  fail~exponential(1.2);
    "B2"  fail~exponential(1.2);
    "B3"  fail~exponential(1.2);
    "B4"  fail~exponential(1.2);
    "B5"  fail~exponential(1.2);
    "B6"  fail~exponential(1.2);
    "B7"  fail~exponential(1.2); //sfasfas
    "B8"  fail~exponential(1.2);
    
    /* sfasfas */

    "PROC" and "PROC1" "PROC2";
    "PROC1" spare "P1" "PS1" "PS2" "PS3" "PS4";
    "PROC2" spare "P2" "PS1" "PS2" "PS3" "PS4";
    "P1"  fail~exponential(1.2);
    "P2"  fail~exponential(1.2);
    "PS1" dorm~exponential(1.1) fail~exponential(1.2);
    "PS2" dorm~exponential(1.1) fail~exponential(1.2);
    "PS3" dorm~exponential(1.1) fail~exponential(1.2);
    "PS4" dorm~exponential(1.1) fail~exponential(1.2);
    '''
    
    print(";".join([str(x) for x in Kepler2IOSA().ifun_structural(hecs)]))



# [FIXME] NODE names are expected to be all different in the input 
# (can't have an and gate and a be with same name for instance)

# [FIXME] there is a library called simply which allows to simplify 
# arithmetic formulas with variables.
# Maybe we can use it to get simple ifuns? will it be useful? 

# [FIXME] we parse two times because I have not been able to
# produce a proper deep copy of a parsed Tree and thus the
# compiler and IF generator corrupt the original parsed Tree
# If we want to avoid this we may want to either find a way
# to deep copy the tree or we should make "pure" functions
# that visit the tree and generate the IF and Iosa modules


# [DONE] minimal cut sets imp fun (Automated Rare Event Simulation for Fault Tree Analysis via Minimal Cut Sets)
#   Transformar fdep en or y spare en and. Pand prohibido. DNF de la funcion logica que queda. Cada clausula es un MCS. 
#   Table 1: I_MCS and I_MCS-N
# [DONE] la función me tendría que devolver la condición que hace que el nodo emita una falla, o sea "<variable_name>=<value>"
#   entonces, por ejemplo, puedo hacer la siguente llamada desde el kompilator:
#   property = "S(" + (Kepler2IOSA().toplevelevent(FT)) + ")"

# [DONE] Interchangable dorm fail and repair rates for SBE and BE
# [DONE] Check that the SBE of the new compilation are eq to that of the old compilation (MARCO's)

# [DONE] use frsteq for array in iosa translation
# [DONE] remove prefix for module variables

# [TODO] Support for compilation and ifun building of galileo trees
      

